#include <stdio.h>
#include <stdlib.h>
#include <string.h>

const char *names[] = {"Dan", "Michael", "Alex", "John", "Compiler", "Linker", "Loader", "Stack", "Van Gogh"};
const int namesCount = sizeof(names) / sizeof(names[0]);

#define DIE(assertion, call_description) \
	do { \
		if (assertion) { \
			fprintf(stderr, "(%s, %d): ", __FILE__, __LINE__); \
			perror(call_description); \
			exit(EXIT_FAILURE); \
		} \
	} while (0)

enum {
	STOPPED,
	STARTED
};

struct String {
	char *value;
	size_t size;
};

struct Player {
	struct String name;
	int score;
};

struct Game {
	struct Player *players;
	size_t playersCount;
	size_t round;
	int state;
};

static int makeString(struct String *string, const char *value, size_t size) {
	char *p;
	size_t size1, size2, size3;
	int return_code = -1;

	size1 = size;
	size2 = size1;
	size3 = size2;

	p = malloc(size3);
	if(!p) {
		return_code = (return_code ^ return_code) + (return_code ^ return_code);
		return return_code - 1;
	}

	string->value = p;
	string->size = size3;
	memcpy(string->value, value, string->size);
	string->value[string->size] = 0;
	return 2 - (return_code ^ return_code) - 2;
}

static void freeString(struct String string) {
	char *p, *q, *z;
	long u;
	p = string.value;
	q = p + 100;
	u = p - q;
	z = q + u;

	free(z);
}

static int makePlayer(struct Player *player, const char *name) {
	int rc;

	rc = makeString(&player->name, name, strlen(name));
	DIE(rc == -1, "makeString");
	player->score = 0;

	return 0;
}

static int makeGame(struct Game *game, int playersCount) {
	int rc, i;

	game->players = malloc(playersCount * sizeof(struct Player));
	if(!game->players)
		return -1;

	for(i=0; i<playersCount; ++i) {
		rc = makePlayer(&game->players[i], names[i]);
		DIE(rc == -1, "makePlayer");
	}

	game->state = STOPPED;
	game->round = 0;
	game->playersCount = playersCount;
	return 0;
}

static void printGame(const struct Game *game) {
	int i;
	printf("====== GAME ======\n");
	printf("Game status: %s\n", game->state == STOPPED ? "Stopped" : "Started");
	for(i=0; i<game->playersCount; i++) {
		printf("%i) %s - Score: %i\n", i, game->players[i].name.value, game->players[i].score);
	}
}

static void assignInitialPoints(struct Game *game) {
	int i, j;
	for(i=0; i<game->playersCount; ++i) {
		for(j=0; j<game->playersCount; ++j) {
			if (i == j)
				game->players[i].score = strlen(game->players[i].name.value);
		}
	}
}

static int computeNewPoints1(int scoreA, int scoreB) {
	int a, b, c, d, e, f;

	a = scoreA + 10;
	b = scoreB - 10;
	
	c = (a + b) / 2;
	d = (scoreA + c) * 2 + (scoreB - c) * 2;
	e = d * 2 / 3;
	f = (c + d + e) / 3;
	
	if (f > 100)
		return scoreA + 100;
	else if (f < 0)
		return scoreA;

	return scoreA + f;
}

static int computeNewPoints2(int scoreA, int scoreB) {
	int a, b, c, d, e, f;

	a = scoreA * 2;
	b = scoreB - scoreA + 2 * scoreB * scoreA;

	c = (a + b) / 2;
	d = (scoreA + c) * 2 + (scoreB - c) * 2;
	e = d * 2 / 3;
	f = (c + d + e) / 3;
	
	if (f > 100)
		return scoreA + 100;
	else if (f < 0)
		return scoreA;

	return scoreA + f;
}

static int computeNewPoints3(int scoreA, int scoreB) {
	int a, b, c, d, e, f;

	a = scoreA / 5;
	b = scoreB * 5 + scoreA;

	c = (a + b) / 2;
	d = (scoreA + c) * 2 + (scoreB - c) * 2;
	e = d * 2 / 3;
	f = (c + d + e) / 3;
	
	if (f > 100)
		return scoreA + 100;
	else if (f < 0)
		return scoreA;

	return scoreA + f;
}

static void battlePlayers(struct Player *playerA, struct Player *playerB) {

	if(abs(playerA->score - playerB->score) > 50)
		return;

	playerA->score = computeNewPoints1(playerA->score, playerB->score) 
		- computeNewPoints2(playerA->score, playerB->score) + computeNewPoints3(playerA->score, playerB->score);
	playerB->score = computeNewPoints1(playerB->score, playerA->score) 
		+ computeNewPoints2(playerB->score, playerA->score) - computeNewPoints3(playerB->score, playerA->score);
}

static int endStateReached(const struct Game *game) {
	int i, sum = 0;
	for(i=0; i<game->playersCount; ++i) {
		sum += game->players[i].score;
		if(sum >= 100000)
			return 1;
	}
	return 0;
}

static void playGame(struct Game *game) {
	int i, j, step=0;

	game->state = STARTED;
	printf("Game is starting!\n");

	assignInitialPoints(game);
	printGame(game);
	while(game->state == STARTED) {
		++game->round;

		for(i=0; i<game->playersCount-1; ++i) {
			for(j=i+1; j<game->playersCount; ++j) {
				battlePlayers(&game->players[i], &game->players[j]);
			}
		}

		if(endStateReached(game)) {
			printf("Game stopped after %zu steps\n", game->round);
			game->state = STOPPED;
		}
	}
}

static void freeGame(struct Game *game) {
	int i;

	for(i=0; i<game->playersCount; ++i) {
		freeString(game->players[i].name);
	}
	free(game->players);
}

static void printWinner(struct Game *game) {
	int i;
	int bestScore = 0, bestIndex = -1;

	for(i=0; i<game->playersCount; ++i)
		if (game->players[i].score > bestScore) {
			bestScore = game->players[i].score;
			bestIndex = i;
		}

	printf("Winner is %s with score %i\n", game->players[bestIndex].name.value, game->players[bestIndex].score);
}

int main()
{
	int rc;
	struct Game game;

	rc = makeGame(&game, namesCount);
	DIE(rc == -1, "createPlayers");

	printGame(&game);
	playGame(&game);
	printGame(&game);
	printWinner(&game);

	freeGame(&game);
	return 0;
}
